package JSPservletPkg;
import java.util.*;
import java.net.*;
import java.io.*;
import java.lang.reflect.*;
import javax.servlet.*;
import javax.servlet.http.*;
import org.osgi.framework.*;
import org.osgi.service.http.*;
import com.sun.jes.service.http.auth.basic.*;

/**
 * It loads classes through ClassEntry with JSPloader and registers them.
 * JES2 activator. Based on Sun SnoopActivator.java
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0
 * @see JSPhandler.ClassEntry
 * @see JSPloader
 */
public class JSPhandler implements BundleActivator, ServiceListener {
  /** path were jars are locally stored - initialized by constructor.
   * @see JSPhandler.ClassEntry
   * @see JSPloader
   */
  String cachePath;
  /** ClassEntry HashMap.
   * key: jar name
   * value: ClassEntry
   */
  HashMap classEntries = new HashMap();
  /** remote locations properties file name - initialized by constructor.
   * @see JSPhandler.ClassEntry
   */
  String remoteLocFile;
  /**
   * Get a File using path parameter and checks it exists or can be created.
   * @param path path to File.
   */
  File getFile(String path) {
    File f = null;
    if (useBundleSpace)
      f = bundleContext.getDataFile(path);
    else
      f = new File(path);
    if (f == null)
      return f;
    if (f.exists())
      return f;
    File parent = f.getParentFile();
    if (!parent.exists())
      parent.mkdirs();
    return f;
  }
  /**
   * Class wraping trace use.
   */
  class Log {
    /** SimpleDateFormat used to build log entries */
    private final java.text.SimpleDateFormat sdf =
      new java.text.SimpleDateFormat("hh:mm:ss:SSS");
    /** PrintWriter initialized by constructor.
     * @see JSPhandler.Log#logprint(String msg)
     */
    private PrintWriter log = null;
    /** true for intensive tracing (INFO) */
    boolean toTrace = false;
    /** Backpointer on JSPhandler. */
    JSPhandler handler;
    /**
     * Constructor.
     * @param jh back pointer on creating JSPhandler.
     */
    Log(JSPhandler jh) {
      handler = jh;
      if (handler.bDisklessClt)
        return;
      try {
        File f = handler.getFile(handler.logfile);
        if (f != null)
          log = new PrintWriter(new BufferedWriter(new FileWriter(f)));
      }
      catch(Exception e) {
        System.out.println(toString() + ".Log " + e);
        log = null;
      }
    }
    /**
     * trace setting method.
     * @param mode new trace mode.
     */
    final synchronized void setTrace(boolean mode) {
      if (log == null && !handler.bDisklessClt)
        return;
      toTrace = mode;
      logprint("trace set " + (mode ? "On":"Off"));
    }
    /**
     * trace getting method.
     * @return current mode.
     */
    final synchronized boolean getTrace() {
      return toTrace;
    }
    /**
     * user logging method.
     * @param   msg message to format and print.
     */
    final synchronized void userprint(String msg) {
      if (handler.bDisklessClt) {
        handler.ih.log("USER  " + sdf.format(new Date()) + " " + msg);
        return;
      }
      if (log == null)
        return;
      log.println("USER  " + sdf.format(new Date()) + " " + msg);
      log.flush();
    }
    /**
     * logging method.
     * @param   msg message to format and print.
     */
    final synchronized void logprint(String msg) {
      if (toTrace && handler.bDisklessClt) {
        handler.ih.log("INFO  " + sdf.format(new Date()) + " " + msg);
        return;
      }
      if (!toTrace)
        return;
      log.println("INFO  " + sdf.format(new Date()) + " " + msg);
      log.flush();
    }
    /**
     * Error logging method.
     * @param   msg message to format and print.
     */
    final synchronized void errorPrint(String msg) {
      if (handler.bDisklessClt) {
        handler.ih.log("ERROR  " + sdf.format(new Date()) + " " + msg);
        return;
      }
      if (log == null)
        return;
      log.println("ERROR " + sdf.format(new Date()) + " " + msg);
      log.flush();
    }
    /**
     * Exception logging method.
     * @param   msg message to format and print.
     */
    final synchronized void exceptionPrint(String msg) {
      if (handler.bDisklessClt) {
        handler.ih.log("EXCEP  " + sdf.format(new Date()) + " " + msg);
        return;
      }
      if (log == null)
        return;
      log.println("EXCEP " + sdf.format(new Date()) + " " + msg);
      log.flush();
    }
    /** Empties the log file. */
    final synchronized void clear() {
      if (handler.bDisklessClt)
        return;
      if (log != null)
        log.close();
      try {
        File f = handler.getFile(handler.logfile);
        if (f != null) {
          if (f.exists())
            f.delete();
          log = new PrintWriter(new BufferedWriter(new FileWriter(f)));
        }
      }
      catch(Exception e) {
        log = null;
      }
      if (log == null)
        toTrace = false;
    }
    /** Close log file. */
    final void close() {
      if (log == null)
        return;
      log.close();
      log = null;
    }
    /** to avoid loitering */
    protected void finalize() {
      close();
    }
  }
  /** Path of log file. */
  String logfile = null;
  /** Initial tracing mode. */
  boolean toTrace = false;
  /** Handle on Log object. */
  Log log;
  /** remote locations Properties - initialized by constructor.
   * @see JSPhandler#update(String jarName, String remoteLocation)
   * @see JSPhandler.ClassEntry
   */
  Properties remoteLocProp = new Properties();
  /** ClassLoader used to create JSPhandler. */
  ClassLoader servletClassLoader;
  /** describes a policy for archive without policy. */
  String defaultPolicy = null;
  /** describes an default policy file and set sandbox for archive execution. */
  String allPermissionPolicy = null;
  /** keystore location. Also states if policy should be downloaded. */
  String keystore = null;
  /** keystore password. */
  String keystorePassword = null;
  /** resources expiration time in seconds - initialized by constructor. */
  int expiration;
  /**
   * URL to access Certificate Revocation List in LDAP.
   * Initialized by constructor.
   */
  String CRLURL = null;
  /**
   * Principal to use to connect to LDAP for CRL.
   * Initialized by constructor.
   */
  String CRLLDAPuser = null;
  /**
   * Password to use to connect to LDAP for CRL.
   * Initialized by constructor.
   */
  String CRLLDAPpasswd = null;
  /**
   * Certificate Revocation List Scan Period in seconds.
   * Initialized by constructor.
   */
  int CRLperiod;
  /**
   * URL to access CA certificate in LDAP.
   * Initialized by constructor.
   */
  String CAURL = null;
  /**
   * Principal to use to connect to LDAP for CA certificate.
   * Initialized by constructor.
   */
  String CALDAPuser = null;
  /**
   * Password to use to connect to LDAP for CA certificate.
   * Initialized by constructor.
   */
  String CALDAPpasswd = null;
  /**
   * URL used for redirection from Web Cache.
   */
  String myurl = null;
  /** user for basic page authentication */
  String basicUser = null;
  /** password for basic page authentication */
  String basicPasswd = null;
  /**
   * User for admin page authentication (ServletLog, DisklessLog,
   * ServletStat, DisklessStat, ServletUpdate).
   */
  String adminUser = null;
  /**
   * Password for admin page authentication (ServletLog, DisklessLog,
   * ServletStat, DisklessStat, ServletUpdate).
   */
  String adminPasswd = null;
  /** record statistics or not */
  boolean toStat = false;
  /** other errors */
  final static int EP_OTHER = 0;
  /** servlet or resource not found */
  final static int EP_NOTFOUND = 1;
  /** servlet attempted an unauthorized access */
  final static int EP_UNAUTHORIZED = 2;
  /** archive certificate revoked */
  final static int EP_REVOKED = 3;
  /** error page array size */
  final static int EP_SIZE = 4;
  /** Scanner wake-up */
  static final int SCANPERIOD = 1000 * 60;
  /** Scanner object */
  Scanner scanner = null;
  /** Area of error handling pages */
  String[] aErrorPage = new String[EP_SIZE];
  /** PageBox unique identifier */
  String ID = null;
  /**
   * Configuration Properties - initialized by constructor.
   * Replaces web.xml that doesn't exist on embedded servers.
   */
  Properties configProp = null;
  /**
   * If true, uses bundle space provided by BundleContext.
   */
  boolean useBundleSpace = true;
  /**
   * If true, diskless and send messages to a diskless server.
   * If false, acts as a diskless server.
   * Default false: writes on local file system.
   */
  boolean bDisklessClt = false;
  /**
   * If true, multicast - use of port and mcg variable.
   * If false, point to point - port contains the listening port.
   *   mcg contains the listening address (interface).
   *   remAddrs is an array of remote server addresses (if bDisklessClt == true).
   *   remPorts is an array of remote server ports (if bDisklessClt == true).
   */
  boolean bMulticast = true;
  /** Server addresses. */
  String[] remAddrs = null;
  /** Server ports. */
  int[] remPorts = null;
  /** Multicast group form xxx.xxx.xxx.xxx or listening address (ptp). */
  String mcg = null;
  /** Multicast port. */
  int port = -1;
  /** Diskless and ICP support. */
  ICPhandler ih = null;
  /** Restores InitParameters from JSPservlet.properties in this jar file. */
  private void initializeParameters() {
    Dictionary dic = bundleContext.getBundle().getHeaders();
    String bundleName = (String)dic.get("Bundle-Name");
    InputStream is = getClass().getResourceAsStream("/" + bundleName +
      ".properties");
    if (is == null) {
      System.out.println("/" + bundleName + ".properties not found");
      return;
    }
    configProp = new Properties();
    try {
      configProp.load(new DataInputStream(is));
    }
    catch(Exception e) {
      System.out.println(toString() +
        ".initializeParameters reading properties:" + e);
      configProp = null;
    }
  }
  /** Parameters retrieval from a property file. */
  String getInitParameter(String parameter) {
    if (configProp == null)
      return null;
    return configProp.getProperty(parameter);
  }
  /**
   * Initializes parameters from property file in archive.
   */
  public void initialize() {
    initializeParameters();
    servletClassLoader = getClass().getClassLoader();
    cachePath = getInitParameter("cachePath");
    if (cachePath == null)
      cachePath = "C:/temp";
    String szToTrace = getInitParameter("toTrace");
    if (szToTrace != null && szToTrace.equalsIgnoreCase("TRUE"))
      toTrace = true;
    logfile = getInitParameter("logfile");
    if (logfile == null)
      logfile = cachePath + "/log.txt";
    String szToStat = getInitParameter("toStat");
    if ((szToStat != null) && (szToStat.equalsIgnoreCase("TRUE")))
      toStat = true;
    String szUseBundleSpace = getInitParameter("useBundleSpace");
    if ((szUseBundleSpace != null) && (szUseBundleSpace.equalsIgnoreCase("FALSE")))
      useBundleSpace = true;
    ID = getInitParameter("ID");
    contextPath = getInitParameter("contextPath");
    remoteLocFile = getInitParameter("remoteLocations");
    if (remoteLocFile == null)
      remoteLocFile = cachePath + contextPath + ".properties";
    defaultPolicy = getInitParameter("defaultPolicy");
    allPermissionPolicy = getInitParameter("allPermissionPolicy");
    String strExpiration = getInitParameter("expiration");
    if ((strExpiration == null) ||
      ((expiration = Integer.parseInt(strExpiration)) == 0))
      expiration = 5;
    keystore = getInitParameter("keystore");
    keystorePassword = getInitParameter("keystorePassword");
    CRLURL = getInitParameter("CRLURL");
    CAURL = getInitParameter("CAURL");
    CRLLDAPuser = getInitParameter("CRLLDAPuser");
    CRLLDAPpasswd = getInitParameter("CRLLDAPpasswd");
    CALDAPuser = getInitParameter("CALDAPuser");
    CALDAPpasswd = getInitParameter("CALDAPpasswd");
    basicUser = getInitParameter("basicUser");
    basicPasswd = getInitParameter("basicPasswd");
    adminUser = getInitParameter("adminUser");
    adminPasswd = getInitParameter("adminPasswd");
    String errorPage = getInitParameter("ErrorPage");
    if (errorPage != null)
      aErrorPage[EP_OTHER] = errorPage;
    errorPage = getInitParameter("NotFoundPage");
    if (errorPage != null)
      aErrorPage[EP_NOTFOUND] = errorPage;
    errorPage = getInitParameter("UnauthorizedPage");
    if (errorPage != null)
      aErrorPage[EP_UNAUTHORIZED] = errorPage;
    errorPage = getInitParameter("RevokedPage");
    if (errorPage != null)
      aErrorPage[EP_REVOKED] = errorPage;
    String strCRLperiod = getInitParameter("CRLperiod");
    if ((strCRLperiod == null) ||
      ((CRLperiod = Integer.parseInt(strCRLperiod)) == 0))
      CRLperiod = 7 * 24 * 3600;
    myurl = getInitParameter("URL");
    if (myurl == null) {
      String myhost = System.getProperty("com.sun.jes.service.http.hostname");
      if (myhost == null) {
        try {
          InetAddress ia = InetAddress.getLocalHost();
          myhost = ia.getHostName();
        }
        catch(UnknownHostException uhe) {}
      }
      if (myhost != null) {
        String myport = System.getProperty("com.sun.jes.service.http.port",
          "8080");
        myurl = "http://" + myhost + ":" + myport;
      }
    }
    String szDisklessClt = getInitParameter("Diskless");
    if ((szDisklessClt != null) && (szDisklessClt.equals("TRUE")))
      bDisklessClt = true;
    mcg = getInitParameter("MulticastGroup");
    String szPort = getInitParameter("MulticastPort");
    if (szPort != null)
      port = Integer.parseInt(szPort);
    String szMulticast = getInitParameter("Multicast");
    if ((szMulticast != null) && (szMulticast.equals("FALSE")))
      bMulticast = false;
    String rema = getInitParameter("RemAddrs");
    if (rema != null) {
      StringTokenizer st = new StringTokenizer(rema, ",");
      remAddrs = new String[st.countTokens()];
      int i = 0;
      while(st.hasMoreTokens())
        remAddrs[i++] = st.nextToken();
    } else
    if (bDisklessClt && !bMulticast)
      System.out.println(toString() +
        ".initialize point to point client without server addresses");
    String remp = getInitParameter("RemPorts");
    if (remp != null) {
      StringTokenizer st = new StringTokenizer(remp, ",");
      int tokenCount = st.countTokens();
      if (tokenCount != remAddrs.length) {
        System.out.println(toString() + ".initialize remAddrs=" + rema +
          " remPorts=" + remp + " invalid");
        remAddrs = null;
      } else {
        remPorts = new int[tokenCount];
        int i = 0;
        while(st.hasMoreTokens())
          remPorts[i++] = Integer.parseInt(st.nextToken());
      }
    } else
    if (bDisklessClt && !bMulticast)
      System.out.println(toString() +
        ".initialize point to point client without server ports");
  }
  /**
   * Implements a scanner thread.
   * Current use: statistics flush.
   */
  class Scanner extends Thread {
    /** JSPhandler back pointer. */
    JSPhandler handler;
    /** flag for termination. */
    boolean toContinue = true;
    /** Constructor. Creates the thread. */
    Scanner(JSPhandler jh) {
      handler = jh;
      this.start();
    }
    /** Checks if Scanner must stop. */
    final synchronized boolean toBeContinued() {
      return toContinue;
    }
    /** Ask Scanner to stop. */
    final synchronized void toEnd() {
      toContinue = false;
    }
    /**
     * Loops to flush Stat entries.
     * Not put log in it for race conditions at termination.
     */
    public void run() {
      while(toBeContinued()) {
        int remaining = SCANPERIOD;
        long nextTime = System.currentTimeMillis() + remaining;
        while(remaining > 0) {
          try {
            sleep(remaining);
          }
          catch(InterruptedException e) {}
          remaining = (int)(nextTime - System.currentTimeMillis());
        }
        if (handler.classEntries == null)
          continue;
        handler.ih.publish();
        if (handler.bDisklessClt)
          handler.ih.update();  // used here to update out of sync servers
        Iterator i = handler.classEntries.values().iterator();
        while(i.hasNext()) {
          ClassEntry ce = (ClassEntry)i.next();
          ce.stat.flush();
        }
      }
    }
  }
  /**
   * Class invoked by JSPhandler for a given jar name.
   * It caches created target servlets and JSPs and manages a JSPloader.
   * The JES2 ClassEntry implements the archive HttpContext.
   */
  class ClassEntry implements HttpContext {
    /** ClassLoader */
    JSPloader jl;
    /** jar name */
    String jarName;
    /** jar remote location (URL) */
    String jarURL;
    /** HashMap
     * key: class name (full path)
     * value: ServletInfo
     * @see JSPhandler.ClassEntry.ServletInfo
     */
    HashMap servletObjects;
    /** back pointer on JSPhandler */
    JSPhandler handler;
    /**
     * Class of servletObjects entries.
     * It wraps Servlet with its protection domain.
     */
    class ServletInfo {
      /** Embedded servlet. */
      Servlet servlet;
      /** Production Domain wrapper around the PD used to define the class. */
      JSPloader.ProtectionDomainInfo pdi;
      /**
       * Constructor.
       * @param   ce ClassEntry creating ServletInfo.
       * @param   className class name of the embedded Servlet.
       * @param   servlet embedded Servlet.
       */
      ServletInfo(ClassEntry ce, String className, Servlet servlet) {
        this.servlet = servlet;
        JSPloader.ClassInfo ci =
          (JSPloader.ClassInfo)ce.jl.classes.get(className);
        pdi = ci.pdi;
      }
      /**
       * check if the credential the class has been defined with are not revoked.
       */
      final boolean check() {
        if (pdi == null)
          return true;
        return pdi.check();
      }
    }
    /**
     * wraps statistic data.
     * uses a Properties to store and persist stats.
     */
    class Stat {
      /** statistic properties */
      Properties prop = null;
      /** ClassEntry back pointer */
      ClassEntry ce = null;
      /** stat file path */
      String statFile = null;
      /**
       * Constructor.
       * @param   ce ClassEntry creating Stat.
       */
      Stat(ClassEntry ce) {
        this.ce = ce;
      }
      /**
       * Statistic initialization.
       * Reads stat property file in memory.
       */
      final synchronized void init() {
        if (ce.handler.bDisklessClt) {
          prop = (Properties)ce.handler.ih.stats.get(ce.jarName);
          if (prop != null) {
            ce.handler.log.logprint(toString() + ".init found existing stats");
            ce.handler.ih.stats.remove(ce.jarName);
          } else {
            ce.handler.log.logprint(toString() + ".init initiated stats");
            prop = new Properties();
          }
          return;
        }
        prop = new Properties();
        statFile = ce.handler.cachePath + "/" + ce.jarName + ".stat";
        File f = ce.handler.getFile(statFile);
        if ((f != null) && f.exists()) {
          try {
            prop.load(new DataInputStream(new FileInputStream(f)));
          }
          catch(Exception e) {
            ce.handler.log.errorPrint(toString() +
              ".init reading properties:" + statFile + " " + e);
          }
        }
        ce.handler.log.logprint(toString() + ".init for " + statFile);
      }
      /**
       * Statistic increment.
       */
      final synchronized void inc(String classpath) {
        if (prop == null)
          return;
        String szCounter = prop.getProperty(classpath);
        if (szCounter == null)
          szCounter = "1";
        else
          szCounter = new Long(Long.parseLong(szCounter) + 1).toString();
        prop.setProperty(classpath, szCounter);
      }
      /**
       * Stores statistics in property file.
       */
      final synchronized void flush() {
        if (prop == null)
          return;
        if (ce.handler.bDisklessClt) {
          ce.handler.ih.statUpdate(ce.jarName, prop);
          return;
        }
        if (statFile == null)
          return;
        try {
          BufferedOutputStream bos = new BufferedOutputStream(
            new FileOutputStream(ce.handler.getFile(statFile)));
          prop.store(bos, null);
          bos.close();
        }
        catch(Exception e) {
          if (ce.handler.scanner.toBeContinued())
            ce.handler.log.errorPrint(toString() + ".flush writing properties:"
              + statFile + " " + e);
        }
      }
      /**
       * Close property file.
       */
      final synchronized void close() {
        if (prop == null)
          return;
        ce.handler.log.logprint(toString() + ".close");
        flush();
        prop.clear();
        prop = null;
      }
      /** to avoid loitering */
      protected void finalize() {
        if (prop == null)
          return;
        close();
      }
    }
    /** Archive statistic object created in ClassEntry constructor. */
    Stat stat = null;
    /** Security (page access) handler object */
    BasicSchemeHandler basic = null;
    /**
     * HttpContext getResource.
     * @param   str resource name.
     */
    public java.net.URL getResource(String str) {
      return null;
    }
    /**
     * HttpContext getMimeType.
     * In most cases, better to return null. Could return "text/html".
     * @param   str resource name.
     */
    public String getMimeType(String str) {
      return null;
    }
    /**
     * Handle security for a request. HttpContext method.
     * HttpService calls this method prior to servicing a request.
     * @param   req servlet request.
     * @param   req servlet response.
     */
    public boolean handleSecurity(HttpServletRequest req,
      HttpServletResponse res) {
      String uri = req.getRequestURI();
      int idx = uri.lastIndexOf('.');
      if (idx != -1)
        return true;  // resource
      String archivePath = handler.contextPath + "/" + jarName + "/";
      if (!uri.startsWith(archivePath))
        return false;
      String servletPath = uri.substring(archivePath.length()).replace('/', '.');
      handler.log.logprint(toString() + ".handleSecurity URI:" + uri +
        " servlet:" + servletPath);
      stat.inc(servletPath);
      if (handler.basicUser == null)
        return true;
      basic = handler.basic;
      if (basic == null)
        return false;
      BasicSchemeHandler.Response response = basic.getResponse(req);
      if (response == null) {
	      challenge(req, res);
        return false;
      }
      String user = response.getName();
      String password = response.getPassword();
      if (handler.basicUser.equals(user)) {
        if (handler.basicPasswd == null)
          return true;
        if (handler.basicPasswd.equals(password))
          return true;
      }
      challenge(req, res);
      return true;
    }
    /**
     * Service method invoked by handleSecurity.
     * @param   req servlet request.
     * @param   req servlet response.
     */
    private final void challenge(HttpServletRequest req,
      HttpServletResponse res) {
      try {
        basic.sendChallenge(res, "dummy");
      }
      catch( IOException ioe ) {
        ioe.printStackTrace();
      }
    }
    /**
     * Constructor.
     * <ul><li>get JAR remote location from remoteLocations.</li>
     * <li>Creates JSPloader.</li>
     * <li>Instanciates a servletObjects target cache.</li></ul><p>
     * @param   jh invoking JSPhandler.
     * @param   jarName application prefix.
     * @param   jarURL jar remote location.
     * @exception JSPloaderException
     *            remote URL of jar not defined.
     */
    ClassEntry(JSPhandler jh, String jarName, String jarURL)
      throws JSPloaderException {
      this.jarName = jarName;
      handler = jh;
      this.jarURL = jarURL;
      stat = new Stat(this);
      if (handler.toStat)
        stat.init();
      buildServlets();
      buildResources();
      handler.log.logprint(toString() + ".ClassEntry(" + jarName + ", " +
        jarURL + ") instantiated");
    }
    /**
     * Checks if a class implements Servlet.
     * Recursive invocation.
     * @param   cl class to check.
     */
    private final boolean isServlet(Class cl) {
      if (cl == null)
        return false;
      Class []cla = cl.getInterfaces();
      int j = 0;
      for (; j < cla.length; ++j) {
        if (cla[j].getName().equals("javax.servlet.Servlet"))
          return true;
        if (isServlet(cla[j]))
          return true;
      }
      if (isServlet(cl.getSuperclass()))
        return true;
      return false;
    }
    /**
     * Creates servlets found in the archive and registers them.
     * Invoked in constructor and update methods.
     * To select classes to instanciate, uses reflection and scans the
     * inheritence hierarchy for Servlet.
     */
    private void buildServlets() throws JSPloaderException {
      jl = new JSPloader(handler, jarName, jarURL);
      servletObjects = new HashMap(jl.getSize());
      Set es = jl.classes.entrySet();
      Iterator i = es.iterator();
      while(i.hasNext()) {
        Map.Entry me = (Map.Entry)i.next();
        String classpath = (String)me.getKey();
        JSPloader.ClassInfo ci = (JSPloader.ClassInfo)me.getValue();
        if (!ci.check())
          continue;
        if (Modifier.isAbstract(ci.cl.getModifiers()))
          continue;
        if (Modifier.isInterface(ci.cl.getModifiers()))
          continue;
        if (!isServlet(ci.cl))
          continue;
        Servlet s = instantiate(classpath, ci.cl);
        try {
          String toRegister = handler.contextPath + "/" + jarName + "/" +
            classpath.replace('.', '/');
          handler.log.logprint(toString() + ".buildServlets register(" +
            toRegister + ", " + s + ")");
          handler.http.registerServlet(toRegister, s, null, this);
        }
        catch(org.osgi.service.http.NamespaceException nse) {
          handler.log.errorPrint(toString() +
            ".buildServlets registerServlet failed " + nse);
        }
        catch(javax.servlet.ServletException se) {
          handler.log.errorPrint(toString() +
            ".buildServlets registerServlet failed " + se);
        }
      }
    }
    /**
     * Register resources starting with the archive name.
     * I use a replacement to com.sun.jes.impl.http.ResourceServlet.
     */
    private void buildResources() {
      try {
        String toRegister = handler.contextPath + "/" + jarName;
        handler.log.logprint(toString() + ".buildResources register(" +
          toRegister + ")");
        handler.http.registerServlet(toRegister, new JSPresourceServlet(this),
          null, this);
      }
      catch(org.osgi.service.http.NamespaceException nse) {
        handler.log.errorPrint(toString() + ".buildResources register failed " +
          nse);
      }
      catch(javax.servlet.ServletException se) {
        handler.log.errorPrint(toString() + ".buildResources register failed " +
          se);
      }
    }
    /**
     * Clears itself and class loaded stuff.
     */
    final void delete() {
      Set s = servletObjects.keySet();
      Iterator i = s.iterator();
      while(i.hasNext()) {
        String classpath = (String)i.next();
        handler.http.unregister(handler.contextPath + "/" + jarName + "/" +
          classpath.replace('.', '/'));
      }
      s = null;
      i = null;
      jl = null;
      servletObjects.clear();
      servletObjects = null;
      System.gc();
    }
    /**
     * Clears itself and class loaded stuff and recreates class loader.
     * <ol>
     * <li>Enumerates servletObjects target cache and destroys each target found
     * </li><li>Clear references.</li>
     * <li>Invokes garbage collector.</li>
     * <li>Creates new JSPloader.</li>
     * <li>Instanciates a new servletObjects target cache.</li></ol><p>
     * @param     remoteLocation URL where to download the jar.
     * @exception JSPloaderException
     *            remote URL of jar not defined.
     * @see JSPhandler.ClassEntry.ServletInfo
     */
    final void update(String remoteLocation) throws JSPloaderException {
      if ((remoteLocation != null) && (remoteLocation.length() > 0))
        jarURL = remoteLocation;
      delete();
      buildServlets();
    }
    /**
     * Creates a servlet using JSPloader and stores it in servletObjects.
     * In JES2 implementation, servletObjects acts as a repository of servlets
     * to unregister.
     * @param     classPath name of the servlet class.
     * @param     jspClass class to instantiate.
     */
    private Servlet instantiate(String classPath, Class jspClass)
      throws JSPloaderException {
      Servlet srv = null;
      try {
        String className = classPath.replace('/', '.');
        srv = (Servlet)jspClass.newInstance();
        if (srv == null)
          throw new JSPloaderException(handler, toString() + ".instantiate(" +
            classPath + ") null object");
        handler.log.logprint(toString() + ".instantiate(" + classPath +
          ") object created");
        servletObjects.put(classPath, new ServletInfo(this, className, srv));
      }
      catch(JSPloaderException jle) {
        throw jle;
      }
      catch(Exception e) {
        throw new JSPloaderException(handler, toString() + ".instantiate(" +
          classPath + ") " + e);
      }
      return srv;
    }
    /** to avoid loitering */
    protected void finalize() {
      stat = null;
      servletObjects.clear();
      servletObjects = null;
      jl = null;
    }
  }
  /**
   * Invoked by ServletUpdate.
   * @param     jarName name of the jar.
   */
  public final synchronized void delete(String jarName) {
    log.logprint(toString() + ".delete(" + jarName + ")");
    try {
      File f = getFile(cachePath + "/" + jarName + ".jar");
      if ((f != null) && (f.exists()))
        f.delete();
      f = getFile(cachePath + "/" + jarName + ".war");
      if ((f != null) && (f.exists()))
        f.delete();
      f = getFile(cachePath + "/" + jarName + ".policy");
      if ((f != null) && (f.exists()))
        f.delete();
    }
    catch(Exception e) {
      log.errorPrint(toString() + ".delete(" + jarName + ") " + e);
    }
    ClassEntry ce = null;
    if (classEntries.containsKey(jarName)) {
      ce = (ClassEntry)classEntries.get(jarName);
      ce.delete();
      classEntries.remove(jarName);
    }
    ih.publish();
    if (remoteLocProp.containsKey(jarName)) {
      remoteLocProp.remove(jarName);
      try {
        BufferedOutputStream bos = new BufferedOutputStream(
          new FileOutputStream(getFile(remoteLocFile)));
        remoteLocProp.store(bos, null);
        bos.close();
      }
      catch(Exception e) {
        log.errorPrint(toString() + ".delete(" + jarName + ") " + e);
      }
    }
  }
  /**
   * Invoked by update.
   * Search ClassEntry with jarName. Creates it if not found.
   * @param     jarName name of the jar.
   * @param     remoteLocation URL where to download the jar.
   * @return    boolean true if jarName found or false otherwise.
   * @exception JSPloaderException in ClassEntry.getProperty.
   * @see JSPhandler.ClassEntry#update(String remoteLocation)
   */
  final boolean memoryUpdate(String jarName, String remoteLocation)
    throws JSPloaderException {
    ClassEntry ce = null;
    if (classEntries.containsKey(jarName)) {
      ce = (ClassEntry)classEntries.get(jarName);
      ce.update(remoteLocation);
      return true;
    }
    if ((remoteLocation == null) || (remoteLocation.length() == 0)) {
      remoteLocation = remoteLocProp.getProperty(jarName);
      if (remoteLocation == null)
        throw new JSPloaderException(this, toString() + ".update(" + jarName +
          ")", JSPloaderException.JLE_UNDEF);
    }
    ce = new ClassEntry(this, jarName, remoteLocation);
    classEntries.put(jarName, ce);
    ih.publish();
    return false;
  }
  /**
   * Invoked by ServletUpdate.
   * <ol><li>Remove local cache</li>
   * <li>If remote location was set by ServletUpdate,
   * updates and persists remote properties.</li></ol><p>
   * @param     jarName name of the jar.
   * @param     remoteLocation URL where to download the jar.
   * @return    boolean true if jarName found or false otherwise.
   * @exception JSPloaderException when persisting modified properties.
   * @see JSPhandler#memoryUpdate(String jarName, String remoteLocation)
   */
  public final synchronized boolean update(String jarName, String remoteLocation)
    throws JSPloaderException {
    log.logprint(toString() + ".update(" + jarName + ", " + remoteLocation + ")");
    if (bDisklessClt) {
      boolean rc = memoryUpdate(jarName, remoteLocation);
      if ((remoteLocation != null) && (remoteLocation.length() > 0))
        remoteLocProp.setProperty(jarName, remoteLocation);
      ih.update();
      return rc;
    }
    try {
      File f = getFile(cachePath + "/" + jarName + ".jar");
      if ((f != null) && (f.exists()))
        f.delete();
      f = getFile(cachePath + "/" + jarName + ".war");
      if ((f != null) && (f.exists()))
        f.delete();
      f = getFile(cachePath + "/" + jarName + ".policy");
      if ((f != null) && (f.exists()))
        f.delete();
    }
    catch(Exception e) {
      log.errorPrint(toString() + ".update(" + jarName + ", " + remoteLocation +
        ") " + e);
    }
    boolean rc = memoryUpdate(jarName, remoteLocation);
    if ((remoteLocation != null) && (remoteLocation.length() > 0)) {
      remoteLocProp.setProperty(jarName, remoteLocation);
      try {
        BufferedOutputStream bos = new BufferedOutputStream(
          new FileOutputStream(getFile(remoteLocFile)));
        remoteLocProp.store(bos, null);
        bos.close();
      }
      catch(Exception e) {
        log.errorPrint(toString() + ".update(" + jarName + ", " + remoteLocation
          + ") " + e);
      }
    }
    return rc;
  }
  /** to avoid loitering */
  protected void finalize() {
    classEntries.clear();
    remoteLocProp.clear();
    classEntries = null;
    remoteLocProp = null;
    log = null;
  }
  /** Handle on ServletLog servlet. */
  ServletLog sl = null;
  /** Handle on DisklessLog servlet. */
  DisklessLog dl = null;
  /** Handle on ServletStat servlet. */
  ServletStat ss = null;
  /** Handle on ServletUpdate servlet. */
  ServletUpdate su = null;
  /**
   * Handle on HttpService. Created in start method.
   * Used to register/unregister servlets.
   */
  HttpService http;
  /**
   * Handle on BasicSchemeHandler. Created in start method.
   * Used in HttpContext to prompt the user for credentials.
   */
  BasicSchemeHandler basic;
  /** Initialized in start from org.osgi.service.http.HttpService */
  ServiceReference httpref;
  /**
   * Initialized in start from
   * com.sun.jes.service.http.auth.basic.BasicSchemeHandler
   */
  ServiceReference httpauthref;
  /** ServletUpdate alias */
  final String SERVLETUPDATE_ALIAS = "/ServletUpdate";
  /** ServletLog alias */
  final String SERVLETLOG_ALIAS = "/ServletLog";
  /** DisklessLog alias */
  final String DISKLESSLOG_ALIAS = "/DisklessLog";
  /** ServletStat alias */
  final String SERVLETSTAT_ALIAS = "/ServletStat";
  /** Used to identify a particular JSPservlet bundle. */
  String contextPath = null;
  /** Context of the bundle (parameter of start method) */
  BundleContext bundleContext;
  /**
   * Invoked by ICPhandler.
   * Initialize remoteLocProp from ICP_OP_ARCHIVE_STATUS
   * @param     is remote property stream.
   */
  String disklessRemotePropInit(InputStream is) {
    try {
      remoteLocProp.load(new DataInputStream(is));
    }
    catch(Exception e) {
      System.out.println(toString() +
        ".disklessRemotePropInit reading properties: " + e);
    }
    if (!toTrace || remoteLocProp.isEmpty())
      return null;
    String nameVal = "";
    Enumeration en = remoteLocProp.keys();
    while(en.hasMoreElements()) {
      String key = (String)en.nextElement();
      nameVal += "\n\t" + key + "=" + remoteLocProp.getProperty(key);
    }
    return nameVal;
  }
  /**
   * BundleActivator method invoked at bundle start.
   * @param     bc context of the bundle.
   * @exception BundleException propagated from framework invocation.
   */
  public void start(BundleContext bc) throws BundleException {
	  bundleContext = bc;
    initialize();
    if (bDisklessClt) {
      ih = new ICPhandler(this);
      log = new Log(this);
      log.setTrace(toTrace);
    } else {
      log = new Log(this);
      log.setTrace(toTrace);
      ih = new ICPhandler(this);
      File f = getFile(remoteLocFile);
      if ((f != null) && f.exists()) {
        try {
          remoteLocProp.load(new DataInputStream(new FileInputStream(f)));
        }
        catch(Exception e) {
          log.errorPrint(toString() + ".JSPhandler reading properties:" +
            remoteLocFile + " " + e);
        }
      }
    }
    scanner = new Scanner(this);
	  sl = new ServletLog(this);
	  dl = new DisklessLog(this);
	  ss = new ServletStat(this);
	  su = new ServletUpdate(this);
	  httpref = bc.getServiceReference("org.osgi.service.http.HttpService");
	  if (httpref != null) {
      http = (HttpService) bc.getService(httpref);
      httpRegister();
    }
    httpauthref = bc.getServiceReference(
      "com.sun.jes.service.http.auth.basic.BasicSchemeHandler");
	  if (httpauthref != null) {
      basic = (BasicSchemeHandler) bc.getService(httpauthref);
    }
	  try {
	    bc.addServiceListener(this,
        "(objectClass=org.osgi.service.http.HttpService)");
	    bc.addServiceListener(this,
		    "(objectClass=com.sun.jes.service.http.auth.basic.BasicSchemeHandler)");
    }
    catch(InvalidSyntaxException ise) {
      ise.printStackTrace();
    }
  }
  /**
   * ServiceListener's serviceChanged method.
   * Note start has subscribed to events from org.osgi.service.http.HttpService
   * and com.sun.jes.service.http.auth.basic.BasicSchemeHandler.
   * @param     event notification of a service change.
   */
  public synchronized void serviceChanged(ServiceEvent event) {
	  if (event.getType() == ServiceEvent.UNREGISTERING) {
	    if (event.getServiceReference().equals(httpref)) {
        httpUnregister();
        httpref = null;
        http = null;
	    } else
      if (event.getServiceReference().equals(httpauthref)) {
		    httpauthref = null;
		    basic = null;
      }
    } else
    if ( event.getType() == ServiceEvent.REGISTERED ) {
      httpref = bundleContext.getServiceReference(
        "org.osgi.service.http.HttpService");
	    if ( httpref != null && http == null ) {
        http = (HttpService) bundleContext.getService(httpref);
        httpRegister();
      }
	    httpauthref = bundleContext.getServiceReference(
        "com.sun.jes.service.http.auth.basic.BasicSchemeHandler");
	    if ( httpauthref != null && basic == null )
		    basic = (BasicSchemeHandler) bundleContext.getService(httpauthref);
    }
  }
  /** Method invoked from start and serviceChanged */
  private void httpRegister() {
    try {
      http.registerServlet(contextPath + SERVLETUPDATE_ALIAS, su, null, su);
      http.registerServlet(contextPath + SERVLETLOG_ALIAS, sl, null, sl);
      http.registerServlet(contextPath + DISKLESSLOG_ALIAS, dl, null, dl);
      http.registerServlet(contextPath + SERVLETSTAT_ALIAS, ss, null, ss);
      Set s = remoteLocProp.entrySet();
      Iterator i = s.iterator();
      while(i.hasNext()) {
        Map.Entry me = (Map.Entry)i.next();
        try {
          ClassEntry ce = new ClassEntry(this, (String)me.getKey(),
            (String)me.getValue());
          classEntries.put(me.getKey(), ce);
        }
        catch(JSPloaderException je) {}
      }
      ih.publish();
    }
    catch(ServletException se) {
      se.printStackTrace();
    }
    catch(NamespaceException nse) {
      nse.printStackTrace();
    }
  }
  /** Method invoked from stop and serviceChanged */
  private void httpUnregister() {
	  try {
      http.unregister(contextPath + SERVLETUPDATE_ALIAS);
      http.unregister(contextPath + SERVLETLOG_ALIAS);
      http.unregister(contextPath + DISKLESSLOG_ALIAS);
      http.unregister(contextPath + SERVLETSTAT_ALIAS);
      Collection c = classEntries.values();
      Iterator i = c.iterator();
      while(i.hasNext()) {
        ClassEntry ce = (ClassEntry)i.next();
        ce.delete();
      }
      classEntries.clear();
      log.close();
      System.gc();
    }
    catch(IllegalArgumentException iae) {}
  }
  /**
   * BundleActivator method invoked at bundle stop.
   * It invokes scanner.toEnd to stop scanner thread.
   * It can take up to SCANPERIOD to end.
   * @param     bc context of the bundle.
   * @exception BundleException propagated from framework invocation.
   */
  public void stop(BundleContext bc) throws BundleException {
    ih.close();
    scanner.toEnd();
    if (http != null) {
	    httpUnregister();
    }
  }
}