package JSPservletPkg;

import java.lang.*;
import java.util.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.util.jar.*;
import java.util.zip.*;
import java.security.*;
import java.security.cert.*;
import java.lang.reflect.*;
import org.osgi.framework.*;
/**
 * ClassLoader invoked by ClassEntry for a given jar name.
 * Version 1.0.2:
 * Use of AccessController.doPrivileged for invocation from sandboxed archives.
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0, 2
 * @see JSPservletPkg.JSPhandler
 * @see JSPservletPkg.JSPhandler.ClassEntry
 */
public class JSPloader extends ClassLoader {
  /** back pointer on JSPhandler */
  JSPhandler handler;
  /** jar remote location (URL) */
  String jarURL;
  /** jar name */
  String jarName;
  /** Loaded classes HashMap.
   * key: full class name
   * value: ClassInfo
   * @see JSPservletPkg.JSPloader.ClassInfo
   */
  HashMap classes = null;
  /** classes size */
  private int size = 0;
  /** Loaded resources HashMap.
   * key: remote location
   * value: byte[]
   * @see JSPservletPkg.JSPloader.ResourceEntry
   */
  HashMap resources;
  /** parent class loader */
  ClassLoader parent;
  /** base URL for policy checking */
  private URL baseURL;
  /** policy set by default */
  private static sun.security.provider.PolicyFile defaultPolicy = null;
  /** application server policy */
  private static sun.security.provider.PolicyFile allPermissionPolicy = null;
  /** policy used by loaded classes */
  private sun.security.provider.PolicyFile policy = null;
  /** policy file used by loaded classes */
  private String policyFile;
  /** part to remove in case of war classes */
  private static final String webInfClasses = "WEB-INF/classes/";
  /** null certificate list (unsigned archive) */
  private static final java.security.cert.Certificate[] certificate0 =
    new java.security.cert.Certificate[0];
  /** ProtectionDomain for classes without certificate */
  private ProtectionDomain protectionDomain0 = null;
  /** map of protection domains */
  private HashMap protectionDomains = new HashMap();
  /** cache file timestamp - used for caching */
  private long lastModified = 0L;
  /** CRLchecker instance */
  private Object crlChecker = null;
  /** CRLchecker check method */
  private Method crlCheckMethod = null;
  /** if true has tried to install jsse */
  private static boolean isSSLset = false;
  /** PageBoxAPI class - used at loadClass */
  private Class pageBoxClass = PageBoxAPI.class;
  /** JES2 jsse service. */
  final private static String SSL_SERVICE = "com.sun.jes.service.ssl.SSLService";
  /** SSL service reference. */
  private static ServiceReference sslSR = null;
  /**
   * Installs jsse.
   * @param   log JSPhandler log.
   */
  private final static synchronized void setSSL(JSPhandler handler) {
    if (isSSLset)
      return;
    sslSR = handler.bundleContext.getServiceReference(SSL_SERVICE);
    if (sslSR == null)
      handler.log.errorPrint("JSPloader.setSSL ssl unusable");
    else
      isSSLset = true;
  }
  /**
   * Constructor. Initialize variables. Can load classes without resolving them.
   * <ol><li>If no sandbox:</li>
   * <ol><li>Tries loading classes from cache (loadClassDataFS).</li>
   * <li>Tries loading classes from remote location (loadClassDataURL).</li></ol>
   * <li>If sandbox:</li>
   * <ol><li>Defines all permission and default policies.</li>
   * <li>Loads archive policy and update keystore if needed.</li>
   * <li>Instantiates CRLchecker.</li>
   * <li>Tries loading classes from cache (loadClassDataFS).</li>
   * <li>Tries copying archive from remote location and load classes.</li></ol>
   * </ol>
   * @param   jh invoking JSPhandler.
   *          Used to access properties, member variables and log.
   * @param   jarName application prefix.
   * @param   jarURL jar remote location (URL).
   */
  public JSPloader(JSPhandler jh, String jarName, String jarURL)
    throws JSPloaderException {
    super();
    handler = jh;
    this.jarURL = jarURL;
    this.jarName = jarName;
    parent = getParent();
    handler.log.logprint(toString() + ".JSPloader(" + jarName + ", " + jarURL +
      ", " + jh.cachePath + ", " + parent + ")");
    setSSL(handler);
    if ((jh.allPermissionPolicy == null) || (jh.defaultPolicy == null)) {
      if (handler.bDisklessClt || !loadClassDataFS()) {
        if (!loadClassDataURL())
          throw new JSPloaderException(handler, toString() +
            ".JSPloader unable to load jar");
      }
      return;
    }
    synchronized(JSPloader.class) {
      // JSPhandler's defaultPolicy and allPermissionPolicy are file names
      // defaultPolicy and allPermissionPolicy are policies
      if (defaultPolicy == null) {
        System.setProperty("java.security.policy", jh.defaultPolicy);
        defaultPolicy = new sun.security.provider.PolicyFile();
        defaultPolicy.refresh();
      }
      if (System.getSecurityManager() == null) {
        System.setProperty("java.security.policy", jh.allPermissionPolicy);
        allPermissionPolicy = new sun.security.provider.PolicyFile();
        allPermissionPolicy.refresh();
        Policy.setPolicy(allPermissionPolicy);
        System.setSecurityManager(new SecurityManager());
      }
      handler.log.logprint(toString() + ".JSPloader CodeSource " +
        JSPloader.class.getProtectionDomain().getCodeSource().getLocation());
      PermissionCollection pc =
        JSPloader.class.getProtectionDomain().getPermissions();
      Enumeration en = pc.elements();
      while(en.hasMoreElements()) {
        Permission per = (Permission)en.nextElement();
        handler.log.logprint(toString() + ".JSPloader " + per);
      }
    }
    if (handler.CRLURL != null) {
      try {
        Class crlCheckerClass = Class.forName("JSPservletPkg.CRLchecker");
        Class[] parm = new Class[] {JSPhandler.class};
        Constructor crlConstr = crlCheckerClass.getConstructor(parm);
        Object[] arg = new Object[] {handler};
        crlChecker = crlConstr.newInstance(arg);
        Class[] parameterTypes = new Class[]{
          java.security.cert.Certificate[].class};
        crlCheckMethod = crlCheckerClass.getMethod("check", parameterTypes);
        parm = new Class[] {JSPloader.class, String.class, String.class};
        RPAconstr = RPAclass.getConstructor(parm);
      }
      catch(Exception e) {
        throw new JSPloaderException(handler, toString() +
          ".JSPloader unable to instanciate CRLchecker " + e);
      }
    }
    try {
      baseURL = new URL(jarURL + "/" + jarName + "/");
    }
    catch(MalformedURLException mue) {
      throw new JSPloaderException(handler, toString() + ".JSPloader " + mue);
    }
    if (handler.bDisklessClt) {
      /* In case of diskless client, keystore is the full path to a preinstalled
       * keystore and policyFile is java.policy in the same dir. */
      if (handler.keystore != null) {
        File keystoreFile = new File(handler.keystore);
        if ((keystoreFile != null) && keystoreFile.exists()) {
          File dir = keystoreFile.getParentFile();
          File policyF = new File(dir, "java.policy");
          if ((policyF != null) && policyF.exists()) {
            synchronized(JSPloader.class) {
              System.setProperty("java.security.policy",
                policyF.getAbsolutePath());
              policy = new sun.security.provider.PolicyFile();
              policy.refresh();
            }
            handler.log.logprint(toString() + ".JSPloader codebase:" + jarURL +
              "/" + jarName + "/ policy:" + policyF.getAbsolutePath());
          }
        }
      }
      if (policy == null)
        policy = defaultPolicy;
      if (!loadClassDataURL())
        throw new JSPloaderException(handler, toString() +
          ".JSPloader unable to load jar");
    } else {
      policyFile = jh.cachePath + "/" + jarName + ".policy";
      File f = handler.getFile(policyFile);
      if ((f == null) || !f.exists()) {
        policyFile = jh.cachePath + "/java.policy";
        f = handler.getFile(policyFile);
      }
      boolean bFound = false;
      if ((f != null) && f.exists()) {
        try {
          // check if it contains an entry for the archive
          BufferedReader br = new BufferedReader(new FileReader(f));
          String ln = null;
          while ((ln = br.readLine()) != null) {
            if (ln.indexOf(jarURL + "/" + jarName + "/") != -1) {
              bFound = true;
              break;
            }
          }
        }
        catch(Exception e) {
          handler.log.errorPrint(toString() + ".JSPloader reading " + policyFile
          + " " + e);
        }
      }
      if (!bFound && (jh.keystore != null))
        policyFile = loadPolicy();
      if (policyFile != null) {
        synchronized(JSPloader.class) {
          if (handler.useBundleSpace) {
            f = handler.bundleContext.getDataFile(policyFile);
            System.setProperty("java.security.policy", f.getAbsolutePath());
          } else
            System.setProperty("java.security.policy", policyFile);
          policy = new sun.security.provider.PolicyFile();
          policy.refresh();
        }
        handler.log.logprint(toString() + ".JSPloader codebase:" + jarURL + "/" +
          jarName + "/ policy:" + policyFile);
      } else
        policy = defaultPolicy;
      if (!loadClassDataFS()) {
        if (!copyClassesURL() || !loadClassDataFS())
          throw new JSPloaderException(handler, toString() +
            ".JSPloader unable to load jar");
      }
    } // disklessSrv
    handler.log.logprint(toString() + ".JSPloader defaultPolicy:" +
      defaultPolicy + " allPermissionPolicy:" + Policy.getPolicy() + " policy:"
      + policy);
    return;
  }
  Constructor RPAconstr = null;
  static Class RPAclass = ResourcePrivilegedAction.class;
  /** Returns classes size. */
  final int getSize() {
    return size;
  }
  /**
   * Returns resource timestamp
   * @param     name name of the resource.
   * @return long either<ol>
   * <li>-1: no information (resource managed by container)</li>
   * <li>0: unknown timestamp</li>
   * <li>>1: valid timestamp</li></ol>
   */
  public synchronized final long getLastModified(String name) {
    if (resources == null)
      return -1L;
    String unslashedName = name;
    if (name.startsWith("/"))
      unslashedName = name.substring(1);
    ResourceEntry re = (ResourceEntry)resources.get(unslashedName);
    long rc = lastModified;
    if ((re != null) && (re.lastModified != -1))
      rc = re.lastModified;
    return rc;
  }
  /**
   * Maintain Protection Domains and CRL scheduled next update.
   */
  class ProtectionDomainInfo {
    /** invoking JSPloader */
    JSPloader jl = null;
    /** embedded ProtectionDomain */
    ProtectionDomain pd = null;
    /** next check time in milliseconds since the epoch */
    long nextCheck = -1;
    /** once a domain is invalid, no need to check again, then set to false */
    boolean isValid = true;
    /** false if Directory server not available */
    boolean isChecked = true;
    /**
     * Constructor.
     * @param     jl invoking JSPloader.
     * @param     pd Protection Domain.
     * @param     nextUpdate next check time in milliseconds since the epoch.
     * @param     isValid boolean false if domain certificate is invalid
     * @param     isChecked boolean false if domain certificate is unchecked
     */
    ProtectionDomainInfo(JSPloader jl, ProtectionDomain pd, long nextUpdate,
      boolean isValid, boolean isChecked) {
      this.isValid = isValid;
      if (!isValid)
        return;
      this.jl = jl;
      this.pd = pd;
      this.isChecked = isChecked;
      long currentTime = System.currentTimeMillis();
      long parmNextCheck = currentTime + (jl.handler.CRLperiod * 1000);
      if (parmNextCheck < nextUpdate)
        nextCheck = parmNextCheck;
      else
        nextCheck = nextUpdate;
      if (jl.handler.log.getTrace())
        jl.handler.log.logprint(toString() + ".ProtectionDomainInfo(" +
          pd.getCodeSource().getLocation() + ", " + new Date(nextUpdate) +
          ") returned " + nextCheck + "(delta=" + (nextCheck - currentTime) +
          ")");
    }
    /**
     * checks if the credential the Protection domain has been defined with
     * are not revoked.
     * @return boolean false if revoked
     */
    final boolean check() {
      if (!isValid)
        return false;
      if ((jl.handler.CRLURL == null) || (jl.crlChecker == null))
        return true;  // uncheckable
      long current = System.currentTimeMillis();
      if (current < nextCheck)
        return true;
      Long LnextUpdate;
      nextCheck = current + (jl.handler.CRLperiod * 1000);
      java.security.cert.Certificate[] certs =
        pd.getCodeSource().getCertificates();
      try {
        Object[] arguments = new Object[] {certs};
        LnextUpdate = (Long)jl.crlCheckMethod.invoke(jl.crlChecker, arguments);
      }
      catch(InvocationTargetException ite) {
        Throwable t = ite.getTargetException();
        if (t instanceof JSPloaderException) {
          JSPloaderException jle = (JSPloaderException)t;
          if ((jle.reason == JSPloaderException.JLE_CERT) ||
            (jle.reason == JSPloaderException.JLE_REVOKED)) {
            isValid = false;
            return false;
          }
        }
        // shouldn't occur
        jl.handler.log.errorPrint(toString() +
          ".check directory access failure for" +
          pd.getCodeSource().getLocation() + " " + t);
        return true;
      }
      catch(IllegalAccessException iae) {
        // shouldn't occur
        jl.handler.log.errorPrint(toString() +
          ".check directory access failure for " +
          pd.getCodeSource().getLocation() + " " + iae);
        return true;
      }
      if (nextCheck > LnextUpdate.longValue())
        nextCheck = LnextUpdate.longValue();
      if (jl.handler.log.getTrace())
        jl.handler.log.logprint(toString() + ".check " +
          pd.getCodeSource().getLocation() + " nextUpdate:" + new Date(nextCheck)
          + " (delta " + (nextCheck - System.currentTimeMillis()) + ")");
      return true;
    }
  }
  /**
   * Maintain class object and serial number.
   * Used to check if its certificate is not revoked.
   * @see JSPservletPkg.JSPloader.ProtectionDomainInfo
   */
  class ClassInfo {
    /** cached class */
    Class cl;
    /** ProtectionDomain information - used to check CRL */
    ProtectionDomainInfo pdi;
    /**
     * Constructor used with a Protection Domain with credentials.
     * @param     cl class object.
     * @param     pdi Protection Domain Information.
     */
    ClassInfo(Class cl, ProtectionDomainInfo pdi) {
      this.cl = cl;
      this.pdi = pdi;
    }
    /**
     * Constructor used with a Protection Domain with credentials for an
     * unallowed class.
     */
    ClassInfo() {
      cl = null;
      pdi = null;
    }
    /**
     * Constructor used without a ProtectionDomain or with
     * a ProtectionDomain without credentials.
     * @param     cl class object.
     */
    ClassInfo(Class cl) {
      this.cl = cl;
      this.pdi = null;
    }
    /**
     * check if the credential the class has been defined with are not revoked.
     * @return boolean true if class is valid<ol>
     */
    final boolean check() {
      if (cl == null)
        return false;
      if (pdi == null)
        return true;
      if (!pdi.check()) {
        cl = null;  // unreference the class - we can no longer use it
        pdi = null;
        return false;
      }
      return true;
    }
  }
  /**
   * Parses the jar stream.
   * <ol><li>Creates local cache if toSave.</li>
   * <li>Scans jis. For each entry, write on local cache if toSave, define the
   * class and save it in classes HashMap.</li></ol>
   * @param     jis JarInputStream from local cache or URL.
   * @param     toSave in case of URL loading, save jar in local cache.
   * @param     isWar true WAR file, false JAR file.
   * @return boolean either<ol>
   * <li>false:  unable to parse the jar file or URL.
   *        Causes: wrong location, network problem</li>
   * <li>true: jar loaded</li></ol>
   */
  private final boolean parseStream(JarInputStream jis, boolean toSave,
    boolean isWar) {
    handler.log.logprint(toString() + ".parseStream InputStream " + jis +
      " to save " + toSave + " war " + isWar);
    JarEntry je = null;
    boolean rc = true;
    try {
      JarOutputStream jos = null;
      String jarPath = null;
      if (toSave) {
        if (isWar)
          jarPath = handler.cachePath + "/" + jarName + ".war";
        else
          jarPath = handler.cachePath + "/" + jarName + ".jar";
        jos = new JarOutputStream(new BufferedOutputStream(
          new FileOutputStream(handler.getFile(jarPath))));
      }
      String k = null;
      while((je = jis.getNextJarEntry()) != null) {
        String entryName = je.getName();
        if (toSave)
          jos.putNextEntry((JarEntry)je.clone());
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        BufferedInputStream bis = new BufferedInputStream(jis);
        int i;
        while((i = bis.read()) != -1)
          baos.write(i);
        byte[] buf = baos.toByteArray();
        if (entryName.endsWith(".class")) {
          if (classes == null)
            classes = new HashMap(100);
          if (isWar)
            k = entryName.substring(webInfClasses.length(),
              entryName.lastIndexOf('.')).replace('/', '.');
          else
            k = entryName.substring(0,
              entryName.lastIndexOf('.')).replace('/', '.');
          ClassInfo jarCl = null;
          if (handler.allPermissionPolicy == null)
            jarCl = new ClassInfo(defineClass(k, buf, 0, buf.length));
          else {
            java.security.cert.Certificate[] c = je.getCertificates();
            if (c == null) {
              if (protectionDomain0 == null) {
                CodeSource cs = new CodeSource(baseURL, certificate0);
                protectionDomain0 = new ProtectionDomain(cs,
                  policy.getPermissions(cs));
                if (handler.log.getTrace()) {
                  PermissionCollection pc = policy.getPermissions(cs);
                  Enumeration en = pc.elements();
                  while(en.hasMoreElements()) {
                    Permission per = (Permission)en.nextElement();
                    handler.log.logprint(toString() + ".parseStream pd0 " + per);
                  }
                }
              }
              jarCl = new ClassInfo(defineClass(k, buf, 0, buf.length,
                protectionDomain0));
            } else {
              ProtectionDomainInfo pdi = null;
              int j = 0;
              Iterator certIt = protectionDomains.keySet().iterator();
              while(certIt.hasNext()) {
                java.security.cert.Certificate[] cert =
                  (java.security.cert.Certificate[])certIt.next();
                if (cert.length != c.length)
                  break;
                for (j = 0; j < cert.length; ++j) {
                  if (!c[j].equals(cert[j]))
                    break;
                }
                if (j == c.length) {
                  pdi = (ProtectionDomainInfo)protectionDomains.get(cert);
                  if (pdi.isValid)
                    jarCl = new ClassInfo(defineClass(k, buf, 0, buf.length,
                      pdi.pd), pdi);
                  else
                    jarCl = new ClassInfo();
                  break;
                }
              }
              if (jarCl == null) {
                boolean isValid = true;
                boolean isChecked = true;
                long nextUpdate = 0;
                if ((handler.CRLURL != null) && (crlChecker != null)) {
                  Object[] arguments = new Object[] {c};
                  try {
                    Long LnextUpdate =
                      (Long)crlCheckMethod.invoke(crlChecker, arguments);
                    nextUpdate = LnextUpdate.longValue();
                  }
                  catch(InvocationTargetException ite) {
                    Throwable t = ite.getTargetException();
                    if (t instanceof JSPloaderException) {
                      JSPloaderException jle = (JSPloaderException)t;
                      if (jle.reason == JSPloaderException.JLE_LDAPACCESS)
                        isChecked = false;
                      else
                      if ((jle.reason == JSPloaderException.JLE_CERT) ||
                        (jle.reason == JSPloaderException.JLE_REVOKED))
                        isValid = false;
                    }
                    handler.log.errorPrint(toString() +
                      ".parseStream directory access failure for " +
                      baseURL + " " + t + " valid:" + isValid + " checked:" +
                      isChecked);
                  }
                  catch(IllegalAccessException iae) {
                    // shouldn't occur
                    handler.log.errorPrint(toString() +
                      ".parseStream directory access failure for " +
                      baseURL + " " + iae);
                  }
                }
                CodeSource cs = new CodeSource(baseURL, c);
                pdi = new ProtectionDomainInfo(this,
                  new ProtectionDomain(cs, policy.getPermissions(cs)),
                  nextUpdate, isValid, isChecked);
                if (handler.log.getTrace()) {
                  PermissionCollection pc = policy.getPermissions(cs);
                  Enumeration en = pc.elements();
                  while(en.hasMoreElements()) {
                    Permission per = (Permission)en.nextElement();
                    handler.log.logprint(toString() + ".parseStream " + per);
                  }
                }
                handler.log.logprint(toString() +
                  ".parseStream ProtectionDomain created for " + baseURL);
                if (isValid)
                  jarCl = new ClassInfo(defineClass(k, buf, 0, buf.length,
                    pdi.pd), pdi);
                else
                  jarCl = new ClassInfo();
                protectionDomains.put(c, pdi);
              }
            }
          }
          if (jarCl != null) {
            classes.put(k, jarCl);
            ++ size;
          }
        } else
        if ((!isWar || !entryName.startsWith("WEB-INF")) &&
          (!entryName.startsWith("META-INF"))) {
          if (resources == null)
            resources = new HashMap();
          ResourceEntry re = new ResourceEntry();
          re.buf = buf;
          re.lastModified = je.getTime();
          resources.put(entryName, re);
          handler.log.logprint(toString() + ".parseStream resource " + entryName
            + " added");
        }
        if (toSave)
          jos.write(buf, 0, buf.length);
        jis.closeEntry();
      }
      jis.close();
      if (toSave) {
        jos.closeEntry();
        jos.close();
        File f = handler.getFile(jarPath);
        if ((f != null) && f.exists())
          lastModified = f.lastModified();
      }
    }
    catch(Exception e) {
      handler.log.errorPrint(toString() + ".parseStream " + e);
      rc = false;
    }
    return rc;
  }
  /**
   * Loads class from local cache (File System).
   * @return boolean either<ol>
   * <li>false:<ul><li>unable to find the jar file</li>
   *        <li>unable to parse the jar file</li>
   *        causes: wrong location</ul>
   * <li>true: jar loaded</li></ol>
   */
  private final boolean loadClassDataFS() {
    String jarPath = handler.cachePath + "/" + jarName;
    handler.log.logprint(toString() + ".loadClassDataFS() " + jarPath);
    JarInputStream jis = null;
    File f = null;
    boolean isWar = true;
    try {
      f = handler.getFile(jarPath + ".war");
      if (f == null)
        return false;
      if (!f.exists()) {
        f = handler.getFile(jarPath + ".jar");
        if ((f == null) || !f.exists())
          return false;
        isWar = false;
      }
      jis = new JarInputStream(new FileInputStream(f));
    }
    catch(Exception e) {
      return false;
    }
    if (f != null)
      lastModified = f.lastModified();
    return parseStream(jis, false, isWar);
  }
  /**
   * Copy classes from remote location (URL).
   * @return boolean either<ol>
   * <li>false:unable to find the jar URL</li>
   * <li>true: jar loaded</li></ol>
   */
  private final boolean copyClassesURL() {
    handler.log.logprint(toString() + ".copyClassesURL() " + jarURL);
    InputStream is = null;
    String suffix = ".war";
    try {
      URL url = new URL(jarURL + "/" + jarName + suffix);
      is = url.openStream();
    }
    catch(Exception e) {}
    if (is == null) {
      suffix = ".jar";
      try {
        URL url = new URL(jarURL + "/" + jarName + suffix);
        is = url.openStream();
      }
      catch(Exception e) {
        handler.log.errorPrint(toString() + ".copyClassesURL() " + jarURL + "/"
          + jarName + ".jar " + e);
      return false;
      }
    }
    if (is == null) {
      handler.log.errorPrint(toString() + ".copyClassesURL() " + jarURL + "/" +
        jarName + ".jar not found");
      return false;
    }
    try {
      BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(
        handler.getFile(handler.cachePath + "/" + jarName + suffix)));
      BufferedInputStream bis = new BufferedInputStream(is);
      int i;
      while((i = bis.read()) != -1)
        bos.write(i);
      bos.close();
      bis.close();
    }
    catch(Exception e) {
      handler.log.errorPrint(toString() + ".copyClassesURL() " + e);
      return false;
    }
    return true;
  }
  /**
   * Copy policy file from remote location (URL).
   * @return String policy file path.
   */
  private final String loadPolicy() {
    handler.log.logprint(toString() + ".loadPolicy() " + jarURL);
    InputStream policyIs = null, certificateIs = null;
    CertificateFactory cf = null;
    try {
      cf = CertificateFactory.getInstance("X.509");
    }
    catch(CertificateException ce) {
      handler.log.errorPrint(toString() +
        ".loadPolicy() unable to get a certificate factory");
      return null;
    }
    try {
      URL url = new URL(jarURL + "/" + jarName + ".policy");
      policyIs = url.openStream();
      url = new URL(jarURL + "/" + jarName + ".cer");
      certificateIs = url.openStream();
    }
    catch(ExceptionInInitializerError eiie) {
      handler.log.errorPrint(toString() + ".loadPolicy() " + eiie + " " +
        eiie.getException() + " CodeSource " +
        JSPloader.class.getProtectionDomain().getCodeSource().getLocation());
      PermissionCollection pc =
        JSPloader.class.getProtectionDomain().getPermissions();
      Enumeration en = pc.elements();
      while(en.hasMoreElements()) {
        Permission per = (Permission)en.nextElement();
        handler.log.logprint(toString() + ".loadPolicy " + per);
      }
      return null;
    }
    catch(Exception e) {
      if (policyIs == null) {
        handler.log.errorPrint(toString() + ".loadPolicy() " + jarURL + "/" +
          jarName + ".policy " + e);
        return null;
      }
      if (certificateIs == null) {
        handler.log.errorPrint(toString() + ".loadPolicy() " + jarURL + "/" +
          jarName + ".cer " + e);
        return null;
      }
    }
    if (policyIs == null) {
      handler.log.errorPrint(toString() + ".loadPolicy() " + jarURL + "/" +
        jarName + ".policy not found");
      return null;
    }
    if (certificateIs == null) {
      handler.log.errorPrint(toString() + ".loadPolicy() " + jarURL + "/" +
        jarName + ".cer not found");
      return null;
    }
    KeyStore ks = null;
    File keystoreFile = handler.getFile(handler.cachePath + "/" +
      handler.keystore);
    try {
      ks = KeyStore.getInstance("JKS", "SUN");
      // update the keystore
      InputStream fis = new FileInputStream(keystoreFile);
      if (handler.keystorePassword == null)
        ks.load(fis, null);
      else
        ks.load(fis, handler.keystorePassword.toCharArray());
      fis.close();
    }
    catch(FileNotFoundException fnfe) {
      // ks should be set if FileNotFoundException
      if (ks == null) {
        handler.log.errorPrint(toString() + ".loadPolicy() null keystore");
        return null;
      }
      handler.log.logprint(toString() + ".loadPolicy keystore " +
        handler.keystore + " doesn't exist");
      try {
        ks.load(null, handler.keystorePassword.toCharArray());
      }
      catch(Exception ce) {
        handler.log.errorPrint(toString() +
          ".loadPolicy keystore initialization " + ce);
        return null;
      }
    }
    catch(Exception e) {
      handler.log.errorPrint(toString() + ".loadPolicy() keystore load " + e);
      return null;
    }
    try {
      OutputStream fos = new FileOutputStream(keystoreFile);
      ks.setCertificateEntry(jarName, cf.generateCertificate(certificateIs));
      if (handler.keystorePassword == null)
        ks.store(fos, null);
      else
        ks.store(fos, handler.keystorePassword.toCharArray());
      fos.close();
      // load the policy
      PrintWriter policyWriter = null;
      File f = handler.getFile(handler.cachePath + "/" + jarName + ".policy");
      if (f.exists())
        f.delete();
      policyWriter = new PrintWriter(new BufferedWriter(new FileWriter(f)));
      policyWriter.println("keystore \"" + handler.keystore + "\";");
      policyWriter.println("grant signedBy \"" + jarName + "\", codeBase \"" +
        jarURL + "/" + jarName + "/\" {");
      BufferedReader br = new BufferedReader(new InputStreamReader(policyIs));
      String ln = null;
      while ((ln = br.readLine()) != null)
        policyWriter.println(ln);
      policyWriter.println("};");
      policyWriter.close();
      br.close();
      return handler.cachePath + "/" + jarName + ".policy";
    }
    catch(Exception e) {
      handler.log.errorPrint(toString() + ".loadPolicy() " + e);
      return null;
    }
  }
  /**
   * Loads class from remote location (URL).
   * @return boolean either<ol>
   * <li>false:<ul><li>unable to find the jar URL</li>
   *        <li>unable to parse the jar URL</li>
   *        causes: wrong location, network problem</ul>
   * <li>true: jar loaded</li></ol>
   */
  private final boolean loadClassDataURL() {
    handler.log.logprint(toString() + ".loadClassDataURL() " + jarURL);
    JarInputStream jis = null;
    boolean isWar = true;
    InputStream is = null;
    try {
      URL url = new URL(jarURL + "/" + jarName + ".war");
      is = url.openStream();
    }
    catch(Exception e) {}
    if (is == null) {
      handler.log.logprint(toString() + ".loadClassDataURL() " + jarURL + "/" +
        jarName + ".war not found");
      try {
        URL url = new URL(jarURL + "/" + jarName + ".jar");
        is = url.openStream();
      }
      catch(Exception e) {
        handler.log.errorPrint(toString() + ".loadClassDataURL() " + jarURL +
          "/" + jarName + ".jar " + e);
        return false;
      }
      if (is == null) {
        handler.log.errorPrint(toString() + ".loadClassDataURL() " + jarURL +
          "/" + jarName + ".jar not found");
        return false;
      }
      isWar = false;
    }
    try {
      jis = new JarInputStream(is);
    }
    catch(Exception e) {
      return false;
    }
    return parseStream(jis, !handler.bDisklessClt, isWar);
  }
  static Class ac = AccessController.class;
  /**
   * Forward load requests. To:
   * <ol><li>Bootstrap and system class loaders (findSystemClass).</li>
   * <li>JSPservlet class loader.</li>
   * <li>Parent class loader.</li></ol><p>
   * @param     name name of the class to load.
   * @return    Class
   */
  private final Class loadForward(String name)
    throws ClassNotFoundException {
    final String classpath = name;
    Class c = null;
    try {
      c = findSystemClass(name);
    }
    catch(Exception e) {}
    if (c != null)
      return c;
    c = (Class)AccessController.doPrivileged(new PrivilegedAction() {
      public Object run() {
        try {
          return handler.servletClassLoader.loadClass(classpath);
        }
        catch(ClassNotFoundException cnfe) {}
        try {
          return parent.loadClass(classpath);
        }
        catch(ClassNotFoundException cnfe) {
          return null;
        }
      }
    });
    if (c == null)
      throw new ClassNotFoundException(toString() + ".loadForward");
    return c;
  }
  /**
   * Externally called method (from ClassEntry and JVM).
   * <ol><li>Calls loadForward for special suffixes.</li>
   * <li>Find the requested class in classes cache.</li>
   * <li>If not found, calls loadForward.</li></ol><p>
   * @param     name name of the class to load.
   * @param     resolve if the class must be resolved.
   * @return    Class
   */
  public synchronized Class loadClass(String name, boolean resolve)
    throws ClassNotFoundException {
    handler.log.logprint(toString() + ".loadClass(" + name + ") " + resolve);
    if (name.equals("JSPservletPkg.PageBoxAPI"))
      return pageBoxClass;
    if (name.equals("javax.servlet.Servlet") || name.startsWith("java.") ||
      name.startsWith("JSPservletPkg."))
      return loadForward(name);
    if ((classes != null) && (classes.containsKey(name))) {
      ClassInfo ci = (ClassInfo)classes.get(name);
      if (!ci.check())
        throw new ClassNotFoundException("Invalidated certificate for " + name);
      if (resolve)
        resolveClass(ci.cl);
      return ci.cl;
    }
    return loadForward(name);
  }
  /**
   * method called by getResourceAsStream.
   * @param     name name of the resource to stream.
   * @return    InputStream
   */
  final private InputStream getResourceForward(String name) {
    handler.log.logprint(toString() + ".getResourceForward(" + name + ")");
    InputStream is = handler.servletClassLoader.getResourceAsStream(name);
    if (is != null)
      return is;
    handler.log.logprint(toString() + ".getResourceForward(" + name +
      ") not found by " + handler.servletClassLoader);
    if (parent != null) {
      is = parent.getResourceAsStream(name);
      if (is == null)
        handler.log.logprint(toString() + ".getResourceForward(" + name +
          ") not found by " + parent);
    }
    return is;
  }
  /**
   * ressources value - maintains a buffer containing the resource and a
   * lastModified time.
   */
  class ResourceEntry {
    /** time in milliseconds since the epoch where the resource was modified */
    long lastModified = -1;
    /** buffer containing the resource */
    byte[] buf = null;
  }
  /** Helper method.
   * ResourceEntry is not in the scope of ResourcePrivilegeAction that needs
   * to create it.
   */
  final ResourceEntry newResourceEntry() {
    return new ResourceEntry();
  }
  /**
   * Externally called method (from Beans.Instantiate or to access resources).
   * This method shouldn't be synchronized as URL handling requires time.
   * @param     name name of the resource to stream.
   * @return    InputStream
   */
  public InputStream getResourceAsStream(String name) {
    handler.log.logprint(toString() + ".getResourceAsStream(" + name + ")");
    String unslashedName = name;
    InputStream is = null;
    if (name.startsWith("/"))
      unslashedName = name.substring(1);
    byte[] buf = null;
    ResourceEntry re = null;
    synchronized(this) {
      if (resources != null) {
        re = (ResourceEntry)resources.get(unslashedName);
        if (re != null)
          buf = re.buf;
        if (buf != null)
          return new ByteArrayInputStream(buf);
      }
    }
    String remoteLoc = jarURL + "/" + unslashedName;
    Object[] arg = new Object[] {this, remoteLoc, name};
    PrivilegedAction pa = null;
    try {
      pa = (PrivilegedAction)RPAconstr.newInstance(arg);
    }
    catch(Exception e) {
      handler.log.logprint(toString() + ".getResourceAsStream " + e);
      return null;
    }
    re = (ResourceEntry)AccessController.doPrivileged(pa);
    if (re != null) {
      synchronized(this) {
        if (resources == null)
          resources = new HashMap();
        resources.put(name, re);
      }
      return new ByteArrayInputStream(re.buf);
    }
    return null;
  }
  /** to avoid loitering */
  protected void finalize() {
    classes.clear();
    resources.clear();
    classes = null;
    resources = null;
    parent = null;
  }
}