package JSPservletPkg;

/**
 * JSPloader specific exception.
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0
 * @see JSPservletPkg.JSPloader
 * @see JSPservletPkg.CRLchecker
 */
public class JSPloaderException extends Exception {
  /** Exception reasons without special handling */
  final static int JLE_OTHER = 0;
  /** Temporary exception: probable cause Directory server unavailability */
  final static int JLE_LDAPACCESS = 1;
  /** Misconfigured LDAP URL: permanent exception */
  final static int JLE_LDAPCONFIG = 2;
  /** Invalid certificate chain: permanent exception for a protection domain */
  final static int JLE_CERT = 3;
  /** Revoked certificate: permanent exception for a protection domain */
  final static int JLE_REVOKED = 4;
  /** archive not defined */
  public final static int JLE_UNDEF = 5;
  /** Exception reason */
  public int reason = JLE_OTHER;
  /**
   * Constructor for reasons without special handling.
   * @param   handler JSPhandler
   * @param   detail diagnosis message
   */
  JSPloaderException(JSPhandler jh, String detail) {
    super(detail);
    jh.log.exceptionPrint(detail);
  }
  /**
   * Constructor for reasons without special handling.
   * @param   handler JSPhandler
   * @param   detail diagnosis message
   * @param   reason exception reason
   */
  JSPloaderException(JSPhandler jh, String detail, int reason) {
    super(detail);
    this.reason = reason;
    if (reason == JLE_LDAPACCESS)
      jh.log.exceptionPrint("LDAP access " + detail);
    else
    if (reason == JLE_LDAPCONFIG)
      jh.log.exceptionPrint("LDAP config " + detail);
    else
    if (reason == JLE_CERT)
      jh.log.exceptionPrint("Invalid cert " + detail);
    else
    if (reason == JLE_UNDEF)
      jh.log.exceptionPrint("Undefined archive " + detail);
    else
    if (reason == JLE_REVOKED)
      jh.log.exceptionPrint("Revoked cert " + detail);
    else
    if (reason == JLE_OTHER)
      jh.log.exceptionPrint(detail);
  }
}
