package JSPservletPkg;
import java.io.*;
import java.net.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import org.osgi.service.http.*;

/**
 * Resource handling servlet.
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0
 * @see JSPhandler.ClassEntry
 * @see JSPloader
 */
public class JSPresourceServlet extends HttpServlet {
  JSPhandler.ClassEntry classEntry;
  /**
   * Constructor.
   * @param     ce ClassEntry that created it.
   */
  public JSPresourceServlet(JSPhandler.ClassEntry ce) {
    classEntry = ce;
  }
  /** Invoked by Servlet container. */
  public void init(ServletConfig config) throws ServletException {
    super.init(config);
  }
  /** SimpleDateFormat used to build RFC conformant strings */
  private static final java.text.SimpleDateFormat sdf =
    new java.text.SimpleDateFormat("EEE, dd MMM yyyy hh:mm:ss");
  /**
   * GET request handling.
   * No exception:
   * <ol><li>The JES2 handling is rude.</li>
   * <li>Exceptions are likely to happen because the response is already
   *    commited. So I silently return.</li></ol>
   * @param     request HttpServletRequest.
   * @param     response HttpServletResponse.
   */
  public void doGet(HttpServletRequest request, HttpServletResponse response) {
    String pathInfo = request.getPathInfo().substring(1);
    ServletContext sc = getServletContext();
    classEntry.handler.log.logprint(toString() + ".doGet(" +
      (sc == null ? pathInfo : sc.getRealPath("/" + pathInfo)) + ") " + response
      + " " + sc);
    try {
      InputStream is = classEntry.jl.getResourceAsStream(pathInfo);
      if (is == null) {
        response.sendError(response.SC_NOT_FOUND);
        return;
      }
      long lastModified = classEntry.jl.getLastModified(pathInfo);
      sdf.setTimeZone(TimeZone.getTimeZone("GMT"));
      String strLastModified = sdf.format(new Date(lastModified)) + " GMT";
      String strETag = "E" + lastModified;
      classEntry.handler.log.logprint(toString() + ".doGet resource(" + pathInfo
        + ") Last-Modified:" + strLastModified);
      String ifNoneMatch = request.getHeader("If-None-Match");
      if (ifNoneMatch != null) {
        classEntry.handler.log.logprint(toString() + ".doGet If-None-Match(" +
          pathInfo + ") ETag:" + strETag + "/" + ifNoneMatch);
        if (ifNoneMatch.equals(strETag)) {
          response.setDateHeader("Expires", System.currentTimeMillis() +
            classEntry.handler.expiration * 1000);
          response.sendError(response.SC_NOT_MODIFIED);
          return;
        }
      }
      String ifModified = request.getHeader("If-Modified-Since");
      if (ifModified != null) {
        classEntry.handler.log.logprint(toString() + ".doGet If-Modified-Since("
          + pathInfo + ") Last-Modified:" + strLastModified + "/" + ifModified);
        if (ifModified.equals(strLastModified)) {
          response.setDateHeader("Expires", System.currentTimeMillis() +
            classEntry.handler.expiration * 1000);
          response.sendError(response.SC_NOT_MODIFIED);
          return;
        }
      }
      // if encoding needed, use javax.mail.internet.MimeUtility
      response.setHeader("Cache-control", "public");
      response.setDateHeader("Date", System.currentTimeMillis());
      response.setHeader("Last-Modified", strLastModified + " GMT");
      response.setHeader("ETag", "E" + lastModified);
      response.setDateHeader("Expires", System.currentTimeMillis() +
        classEntry.handler.expiration * 1000);
      if (sc != null)
        response.setContentType(sc.getMimeType(pathInfo));
      BufferedInputStream bis = new BufferedInputStream(is);
      OutputStream os = response.getOutputStream();
      BufferedOutputStream bos = new BufferedOutputStream(os);
      int i;
      while((i = bis.read()) != -1)
        bos.write(i);
      bos.flush();
      bis.close();
    }
    catch (IllegalStateException ise) {
      classEntry.handler.log.errorPrint(toString() + ".doGet(" + pathInfo + ") "
        + ise);
    }
    catch(IOException ioe) {
      classEntry.handler.log.errorPrint(toString() + ".doGet(" + pathInfo + ") "
        + ioe);
    }
  }
  /**
   * Service method for servlet using RequestDispatcher.
   * Its use is optional. It returns the jar name part of the path.
   * We recommend however its use as the servlet developer should not be aware
   * of its packaging.<p>
   * @param     cl ClassLoader.
   * @return    the jar string.
   */
  public static final String getJAR(ClassLoader cl) {
    if (cl instanceof JSPloader) {
      JSPloader jl = (JSPloader)cl;
      return jl.handler.contextPath + "/" + jl.jarName;
    }
    return "";
  }
}
