package JSPservletPkg;

import java.lang.*;
import java.io.*;
import java.net.*;
import java.security.*;
/**
 * Wraps a resource download.
 * Must be invoked by AccessController.doPrivileged.
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0, 2
 */
class ResourcePrivilegedAction implements PrivilegedAction {
  /** Back pointer on loader. */
  JSPloader loader;
  /** Back pointer on handler. */
  JSPhandler handler;
  /** Resource base URL. */
  String remoteLoc;
  /** Resource path. */
  String name;
  /**
   * Constructor.
   * @param     jl invoking JSPloader.
   * @param     rl resource base URL.
   * @param     name resource path.
   */
  public ResourcePrivilegedAction(JSPloader jl, String rl, String name) {
    loader = jl;
    handler = jl.handler;
    remoteLoc = rl;
    this.name = name;
  }
  /**
   * reads a stream in a byte buffer. Called by run.
   * @param     is input stream.
   */
  final private byte[] getStream(InputStream is) throws IOException {
    BufferedInputStream bis = new BufferedInputStream(is);
      int i;
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      while((i = bis.read()) != -1)
        baos.write(i);
      return baos.toByteArray();
    }
  /** Invoked by AccessController.doPrivileged. */
  public Object run() {
    handler.log.logprint(toString() + ".getResourceAsStream remote location " +
      remoteLoc);
    JSPloader.ResourceEntry re = loader.newResourceEntry();
    try {
      URL url = new URL(remoteLoc);
      URLConnection uc = url.openConnection();
      if (url.getProtocol().equals("http")) {
        HttpURLConnection huc = (HttpURLConnection)uc;
        if (huc.getResponseCode() == HttpURLConnection.HTTP_OK) {
          re.lastModified = uc.getLastModified();
          if (re.lastModified == 0)
            re.lastModified = -1;
          re.buf = getStream(uc.getInputStream());
          return re;
        }
      } else {
        re.lastModified = uc.getLastModified();
        if (re.lastModified == 0)
          re.lastModified = -1;
        re.buf = getStream(uc.getInputStream());
        return re;
      }
    }
    catch(Exception e) {
      handler.log.errorPrint(toString() + ".run " + remoteLoc + " " + e);
    }
    handler.log.logprint(toString() + ".run forward " + name);
    InputStream is = handler.servletClassLoader.getResourceAsStream(name);
    if (is != null) {
      try {
        re.buf = getStream(is);
        re.lastModified = -1;
        return re;
      }
      catch(IOException ioe) {}
    }
    handler.log.logprint(toString() + ".run " + name + " not found by " +
      handler.servletClassLoader);
    if (loader.parent != null)
      is = loader.parent.getResourceAsStream(name);
    if (is != null) {
      try {
        re.buf = getStream(is);
        re.lastModified = -1;
        return re;
      }
      catch(IOException ioe) {}
    }
    handler.log.errorPrint(toString() + ".run " + name + " not found by " +
      loader.parent);
    return null;
  }
}
