package JSPservletPkg;

import javax.servlet.*;
import javax.servlet.http.*;
import java.io.*;
import java.util.*;
import org.osgi.service.http.*;
import com.sun.jes.service.http.auth.basic.*;
/**
 * Update servlet. Equivalent to JSPupdate.
 * Changes:<ul>
 * <li>Version 1.0.1: delete method</li></ul>
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0, 1
 */
public class ServletUpdate extends HttpServlet implements HttpContext {
  /** Invoked by Servlet container. */
  public void init(ServletConfig config) throws ServletException {
    super.init(config);
  }
  /** HTML prefix of error and notification messages */
  final static String msgHeader = "<p><font color=#DB1260 size=4><i>";
  /** HTML suffix of error and notification messages */
  final static String msgTrailer = "</i></font>";
  /** message ArrayList. */
  ArrayList errorMsgs = new ArrayList();
  /**
   * GET request handling.
   * Retrieve jarName parameter and invokes JSPhandler update method.
   * @param     request HttpServletRequest.
   * @param     response HttpServletResponse.
   */
  public void doGet(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {
    errorMsgs.clear();
    String submitVal = request.getParameter("Submit");
    String deleteVal = request.getParameter("Delete");
    boolean askForCreate = false;
    if (handler.classEntries.isEmpty())
      askForCreate = true;
    if (request.getParameter("GC") != null)
      System.gc();
    else
    if ((submitVal != null) || (deleteVal != null)) {
      String jarName = request.getParameter("jarName");
      if (jarName == null || jarName.length() == 0)
        errorMsgs.add("No JAR name supplied!");
      else {
        String withoutExt = jarName;
        int index = jarName.lastIndexOf('.');
        if (index != -1)
      	  withoutExt = jarName.substring(0, index);
        try {
          if (deleteVal != null) {
            handler.delete(withoutExt);
            errorMsgs.add("JAR " + withoutExt + " deleted!");
          } else
          if (handler.update(withoutExt, request.getParameter("remoteLocation")))
            errorMsgs.add("JAR " + withoutExt + " updated!");
          else
            errorMsgs.add("JAR " + withoutExt + " loaded!");
        }
        catch(JSPloaderException jle) {
          if (jle.reason == JSPloaderException.JLE_UNDEF)
            errorMsgs.add("JAR " + withoutExt +
              " not defined - set the remote location!");
          else
            errorMsgs.add("JAR " + withoutExt + " cannot be accessed!");
        }
      }
    }
    response.setContentType("text/html");
    PrintWriter out = new PrintWriter (response.getOutputStream());
    out.println("<html>");
    out.println("<head><title>ServletUpdate</title></head>");
    out.println("<body>");
    out.println("<h2><font color=#DB1260>Servlet update</font></h2><hr>");
    out.println("<form  action=ServletUpdate method=\"GET\">");
    out.println("<table border=0 cellspacing=2 cellpadding=2 width=80%><tr>");
    out.println(
"<td width=80%><font face=\"Helvetica\"><b>Loaded/</b>Defined</td></tr><tr>");
    out.println("<td width=15%><font face=\"Helvetica\"><b>Handler</b></td>");
    out.println("<td width=5%><font face=\"Helvetica\"><b>Trace</b></td>");
    out.println(
      "<td width=40%><font face=\"Helvetica\"><b>Download location</b></td>");
    out.println(
      "<td width=20%><font face=\"Helvetica\"><b>Archive</b></td>");
    out.println("<td width=10%><font face=\"Helvetica\">&nbsp;</td>");
    out.println("<td width=10%><font face=\"Helvetica\">&nbsp;</td></tr>");
    boolean isFirst = true;
    if (request.getParameterNames().hasMoreElements() &&
      ((submitVal == null) || submitVal.equals("Update")) &&
      ((deleteVal == null) || deleteVal.equals("Delete"))) {
      if (request.getParameter("trace" + handler.contextPath) != null)
        handler.log.setTrace(true);
      else
        handler.log.setTrace(false);
    }
    Set definedArchives = handler.remoteLocProp.keySet();
    Iterator defArchIt = definedArchives.iterator();
    while(defArchIt.hasNext()) {
      String archive = (String)defArchIt.next();
      if (request.getParameter("Upd" + handler.contextPath + ":" + archive) !=
        null) {
        try {
          handler.update(archive, null);
          break;
        }
        catch(JSPloaderException jle) {
          errorMsgs.add("update " + handler.contextPath + "/" + archive + " " +
            jle);
          break;
        }
      } else
      if (request.getParameter("Del" + handler.contextPath + ":" + archive) !=
        null) {
        handler.delete(archive);
        break;  // only one delete because ConcurrentModificationException
      }
    }
    Set ces = handler.classEntries.keySet();
    Iterator ceit = ces.iterator();
    ArrayList loadedArchives = new ArrayList();
    while(ceit.hasNext()) {
      String archive = (String)ceit.next();
      if (isFirst) {
        out.println("<tr><td width=15%><font face=\"Helvetica\"><b>" +
          handler.contextPath + "</b></td>");
        out.println(
          "<td width=5%><font face=\"Helvetica\"><input type=\"checkbox\" " +
          (handler.log.getTrace() ? "checked" : "") + " name=\"trace" +
          handler.contextPath + "\" value=on></td>");
        isFirst = false;
      } else {
    out.println("<tr><td width=15%><font face=\"Helvetica\"><b>&nbsp;</b></td>");
    out.println("<td width=5%><font face=\"Helvetica\"><b>&nbsp;</b></td>");
      }
      out.println("<td width=40%><font face=\"Helvetica\"><b>" +
        (String)handler.remoteLocProp.getProperty(archive) + "</b></td>");
      out.println("<td width=20%><font face=\"Helvetica\"><b>" + archive +
        "</b></td>");
      out.println(
"<td width=10%><font face=\"Helvetica\"><input type=\"Submit\" value=\"Upd\" name=\"Upd"
        + handler.contextPath + ":" + archive + "\"></td>");
      out.println(
"<td width=10%><font face=\"Helvetica\"><input type=\"Submit\" value=\"Del\" name=\"Del"
        + handler.contextPath + ":" + archive + "\"></td></tr>");
      loadedArchives.add(archive);
    }
    Object [] loadedArchiveA = loadedArchives.toArray();
    defArchIt = definedArchives.iterator();
    while(defArchIt.hasNext()) {
      String archive = (String)defArchIt.next();
      int i = 0;
      for (; i < loadedArchiveA.length; ++i) {
        if (archive.equals((String)loadedArchiveA[i]))
          break;
      }
      if (i < loadedArchiveA.length)
        continue;
      if (isFirst) {
        out.println("<tr><td width=15%><font face=\"Helvetica\"><b>" +
          handler.contextPath + "</b></td>");
        out.println(
          "<td width=5%><font face=\"Helvetica\"><input type=\"checkbox\" " +
          (handler.log.getTrace() ? "checked" : "") + " name=\"trace" +
          handler.contextPath + "\" value=on></td>");
        isFirst = false;
      } else {
    out.println("<tr><td width=15%><font face=\"Helvetica\"><b>&nbsp;</b></td>");
    out.println("<td width=5%><font face=\"Helvetica\"><b>&nbsp;</b></td>");
      }
      out.println("<td width=40%><font face=\"Helvetica\">" +
        (String)handler.remoteLocProp.getProperty(archive) + "</td>");
      out.println("<td width=20%><font face=\"Helvetica\">" + archive + "</td>");
      out.println(
"<td width=10%><font face=\"Helvetica\"><input type=\"Submit\" value=\"Load\" name=\"Upd"
        + handler.contextPath + ":" + archive + "\"></td>");
      out.println(
"<td width=10%><font face=\"Helvetica\"><input type=\"Submit\" value=\"Del\" name=\"Del"
        + handler.contextPath + ":" + archive + "\"></td></tr>");
    }
    out.println(
"<tr><td width=100%><font face=\"Helvetica\"><b>Definition</b></td></tr><tr>");
    out.println("<td width=20%><font face=\"Helvetica\"><b>JAR Name :</b></td>");
    out.println(
"<td width=20%><font face=\"Helvetica\"><input type=\"text\" name=\"jarName\"></font></td>");
    out.println("<td width=60%><font face=\"Helvetica\"><b>&nbsp;</b></td>");
    out.println(
"</tr><tr><td width=20%><font face=\"Helvetica\"><b>Remote Location :</b></td>");
    out.println(
"<td width=20%><font face=\"Helvetica\"><input type=\"text\" name=\"remoteLocation\"></font></td>");
    out.println("<td width=60%><font face=\"Helvetica\"><b>&nbsp;</b></td>");
    if (askForCreate) {
      out.println(
"</tr><tr><td width=20%><font face=\"Helvetica\"><input type=\"Submit\" value=\"Create\" name=\"Submit\"></td>");
      out.println(
"<td width=20%><font face=\"Helvetica\"><input type=\"Submit\" value=\"Remove\" name=\"Delete\"></td>");
    } else {
      out.println(
"</tr><tr><td width=20%><font face=\"Helvetica\"><input type=\"Submit\" value=\"Update\" name=\"Submit\"></td>");
      out.println(
"<td width=20%><font face=\"Helvetica\"><input type=\"Submit\" value=\"Delete\" name=\"Delete\"></td>");
      out.println(
"<td width=20%><font face=\"Helvetica\"><input type=\"Submit\" value=\"GC\" name=\"GC\"></td>");
    }
    out.println("</tr></form></table><hr>Alexis Grandemange (c)2000 GNU GPL 2");
    Object[] msgA = errorMsgs.toArray();
    for (int i = 0; i < msgA.length; ++i)
      out.println(msgHeader + msgA[i] + msgTrailer);
    out.println("</body></html>");
    out.close();
  }
  public String getServletInfo() {
    return "JSPservletPkg.ServletUpdate Information";
  }
  /**
   * JSPhandler that created ServletUpdate.
   * It implements also BundleActivator, ServiceListener.
   */
  JSPhandler handler;
  /** Page access security. */
  BasicSchemeHandler basic;
  /**
   * Constructor.
   * @param     jh JSPhandler that created it.
   */
  public ServletUpdate(JSPhandler jh) {
	  handler = jh;
  }
  /**
   * HttpContext interface getResource.
   * @param     str requested resource.
   */
  public java.net.URL getResource(String str) {
	  return null;
  }
  /**
   * HttpContext interface's getMimeType.
   * @param     str requested resource.
   */
  public String getMimeType(String str) {
	  return null;
  }
  /**
   * HttpContext interface's handleSecurity.
   * @param     request HttpServletRequest.
   * @param     response HttpServletResponse.
   */
  public boolean handleSecurity(HttpServletRequest req,
    HttpServletResponse res) {
    if (handler.adminUser == null)
      return true;
	  basic = handler.basic;
	  if (basic == null)
      return false;
    BasicSchemeHandler.Response response = basic.getResponse(req);
    if (response == null) {
	    challenge(req, res);
	    return false;
    }
	  String user = response.getName();
	  String password = response.getPassword();
    if ((handler.adminPasswd == null) && handler.adminUser.equals(user))
      return true;
	  if (handler.adminUser.equals(user) && handler.adminPasswd.equals(password))
      return true;
    challenge(req, res);
	  return false;
  }
  /**
   * Invokes BasicSchemeHandler sendChallenge.
   * @param     request HttpServletRequest.
   * @param     response HttpServletResponse.
   */
  final void challenge(HttpServletRequest req, HttpServletResponse res) {
    try {
		  basic.sendChallenge(res, "dummy");
    }
    catch( IOException ioe ) {
		  ioe.printStackTrace();
    }
  }
}
